# -*- coding: utf-8 -*-

# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENSE.TXT

"""
This module contains classes for X-Ray internal objects
"""
import hashlib
import logging
import os
from datetime import datetime, timedelta
from fnmatch import fnmatchcase
from typing import Iterator, List, Optional, NamedTuple
from urllib.parse import urlparse

from .constants import request_data_storage, api_server
from .exceptions import XRayManagerError
from .utils import (
    timestamp,
    skeleton_update,
    safe_move,
    dbm_storage_update,
    _cagefsctl_get_prefix,
    _selectorctl_get_version,
    _is_selector_phpd_location_set,
    _is_cagefs_enabled
)

logger = logging.getLogger('types')


class URL(NamedTuple):
    """
    Representation of an URL
    """
    domain_name: str
    uri_path: str


class ContinuousTask(NamedTuple):
    """
    Representation of a continuous tracing task
    """
    creation_time: int
    domain: str
    original_url: str
    email: str
    execution_count: int
    status: int


def url_split(url: str) -> URL:
    """
    Split URL into domain_name and uripath including query string
    :param url: URL of format protocol://domain/path;parameters?query#fragment
    :return: namedtuple URL(domain_name, uripath)
    """
    fragments = urlparse(url)
    qs = f'?{fragments.query}' if fragments.query else ''
    uri = f'{fragments.path}{qs}' if fragments.path else '/'
    _no_www_netloc = fragments.netloc.replace('www.', '')
    _no_port_netloc = _no_www_netloc.split(':')[0]
    logger.info('Parsed %s into %s:%s', url, _no_port_netloc, uri)
    return URL(_no_port_netloc, uri)


class DomainInfo:
    """
    Simple container class for storing information about domain.
    Allows to add whatever attributes.
    Implements two properties:
     - version set in CL selector
     - direct ini path in CageFS
    """

    def __init__(self, **kwargs):
        self.name = None
        self.user = None
        self.panel_php_version = None
        self.saved_selector_php_version = 'native'
        self.panel_fpm = None
        for k, v in kwargs.items():
            self.__setattr__(k, v)

    def __repr__(self):
        return str(self.__dict__)

    def __str__(self):
        return f'Domain {self.name}'

    def add_cagefs_dirpath(self, basename: str) -> Optional[str]:
        """
        Get path to ini location directory in CageFS
        """
        prefix = _cagefsctl_get_prefix(self.user)
        if prefix is None:
            raise ValueError(
                'CageFS prefix resolved as None, but should be a number')
        return f'/var/cagefs/{prefix}/{self.user}/etc/cl.php.d/{basename}'

    @property
    def selector_php_version(self) -> Optional[str]:
        """
        Get PHP version set in selector
        :param domain_user: user of domain
        :return: PHP version in format of 'alt-phpXY'
        """
        if self.saved_selector_php_version == 'native':
            result = _selectorctl_get_version(self.user)
            if result is not None:
                out, err = result
                not_err_result = 'ERROR' not in err and 'ERROR' not in out
                if not_err_result and 'native' not in out:
                    short_version = out.split()[0].strip()
                    self.saved_selector_php_version = f"alt-php{''.join(short_version.split('.'))}"
                else:
                    # selectorctl success, but returned ERROR or native version
                    self.saved_selector_php_version = None
            else:
                # selectorctl command failed or just returned non-zero retcode
                self.saved_selector_php_version = None
        return self.saved_selector_php_version

    @property
    def selector_ini_path(self) -> Optional[str]:
        """
        Resolve direct ini path in CageFS
        """
        return self.add_cagefs_dirpath(self.selector_php_version)

    @property
    def phpd_location_ini_path(self) -> Optional[str]:
        """
        Resolve direct ini path for the case of php.d.location = selector.
        This, if set, affects only alt-php versions with enabled CageFS.
        Also, consider that FPM is not compatible with selector behaviour
        """
        alt_php_in_cage = 'alt-php' in self.panel_php_version and _is_cagefs_enabled(self.user)
        if _is_selector_phpd_location_set() and alt_php_in_cage and not self.panel_fpm:
            return self.add_cagefs_dirpath(self.panel_php_version)


class Task:
    """
    Container class for storing task data
    """
    cron_file = '/etc/cron.d/xray-manager'

    def __init__(self, *, url: str, client_ip: str, tracing_by: str,
                 tracing_count: int, ini_location: str = 'unknown',
                 task_id: str = 'unknown', status: str = 'unknown',
                 starttime: int = -1, initial_count: int = 0,
                 request_count: int = 0, auto_task: bool = False,
                 user: str = None):
        self.url = url
        self.client_ip = client_ip
        self.tracing_by = tracing_by
        self.tracing_count = tracing_count
        self.ini_location = ini_location
        self.task_id = task_id
        self.status = status
        self.starttime = starttime
        self.request_count = request_count
        self.initial_count = tracing_count if initial_count == 0 else initial_count
        self.auto_task = auto_task
        self.user = user
        logger.info('Instantiate Task %s: %s',
                    self.task_id, self.as_dict())

    def __repr__(self):
        return str(self.__dict__)

    def __str__(self):
        return f'Task {self.task_id}'

    @property
    def fake_id(self) -> str:
        return hashlib.blake2b(self.task_id.encode(),
                               digest_size=10).hexdigest()

    @property
    def shared_link(self) -> str:
        if self.auto_task:
            return f'https://{api_server}/static/xray/reports/?tracing_task_id={self.task_id}'
        raise XRayManagerError(
            'No shared link could be obtained for manual task')

    def as_dict(self) -> dict:
        """
        Represent task as a regular dictionary without task_id field
        :return: dict
        """
        return {k: v for k, v in self.__dict__.items() if k != 'task_id'}

    def generate(self) -> str:
        """
        Generate task for writing into xray.ini file.
        :return: domain_name:uripath:client_ip:tracing_task_id
        """
        url_data = ':'.join(url_split(self.url))
        task_view = ':'.join((url_data, self.client_ip, self.fake_id))
        logger.info('Generated task %s', task_view)
        return task_view

    def ini_file(self) -> str:
        """
        Full path to xray.ini file
        :return: xray.ini full path
        """
        return os.path.join(self.ini_location, 'xray.ini')

    def set_cronjob(self, system_id: str) -> None:
        """
        Set a cron job for stop task by time
        :param system_id: a unique system ID
        """
        if self.tracing_by == 'time':
            stop_at = datetime.fromtimestamp(self.starttime) + timedelta(
                minutes=self.tracing_count + 1)
            shceduled = f'{stop_at.minute} {stop_at.hour} {stop_at.day} {stop_at.month} *'
            cmd = f'root /usr/sbin/cloudlinux-xray-manager stop'
            params = f'--system_id {system_id} --tracing_task_id {self.task_id}'
            cron_line = f'{shceduled}\t{cmd} {params}\n'
            individual_cron_file = self.cron_file + f'_{self.task_id}'
            logger.info('Generating cron file %s with command %s',
                        individual_cron_file, cron_line)
            with open(individual_cron_file, 'w') as cron:
                cron.write(cron_line)

    def drop_cronjob(self) -> None:
        """
        Drop existing cron job file for current task
        """
        job = self.cron_file + f'_{self.task_id}'
        logger.info('Removing cron job %s', job)
        try:
            os.unlink(job)
        except FileNotFoundError:
            logger.info('No such job')

    def read_ini(self) -> List[str]:
        """
        Read contents of xray.ini
        Raise XRayError in case if xray.ini does not exists
        :return: list of lines contained in xray.ini
        """
        if os.path.exists(self.ini_file()):
            with open(self.ini_file()) as existing:
                return existing.readlines()
        else:
            raise XRayManagerError('Failed to find xray.ini')

    @staticmethod
    def ini_tasks_parsed(all_tasks: str) -> Iterator[URL]:
        """
        Iterate over URL of each task in xray.tasks=* line
        :return: list of URLs
        """
        for task in all_tasks[11:].split(','):
            info = task.split(':')
            yield URL(info[0], info[1])

    def is_a_duplicate(self):
        """
        Check if task is duplicated by another tasks
        already present in current xray.ini
        :return: True if task is a duplicate, False otherwise
        """
        domain, uri = url_split(self.url)
        try:
            existing_tasks = [l for l in self.read_ini() if 'xray.tasks' in l][
                0]
        except XRayManagerError:
            return False

        def check(t):
            """
            Match domain name and uri separately, but return joined result
            :param t: task
            :return: True if both domain and uri matches task, False otherwise
            """
            direct_match = domain == t.domain_name and uri == t.uri_path
            wildcard_match = fnmatchcase(domain,
                                         t.domain_name) and fnmatchcase(uri,
                                                                        t.uri_path)
            return direct_match or wildcard_match

        return any(
            [check(task) for task in self.ini_tasks_parsed(existing_tasks)])

    @staticmethod
    def so_path(php_ver: str = None) -> str:
        """
        Ensure full path for versions, which are not among known.
        In such a case version is expected as two numbers, e.g. 56 or 74
        """
        if php_ver is None or len(php_ver) > 2:
            return 'xray.so'
        return f'/opt/alt/php{php_ver}/usr/lib64/php/modules/xray.so'

    def _add(self, version: str = None) -> None:
        """
        Add given task into xray.ini file
        """
        working_ini_file = self.ini_file() + '.tmp'
        try:
            existing_contents = self.read_ini()
        except XRayManagerError:
            ini_contents = f"""extension={self.so_path(version)}
xray.tasks={self.generate()}
"""
        else:
            def update():
                """
                A generator function aimed to update xray.task field
                """
                for line in existing_contents:
                    yield f'{line.strip()},{self.generate()}\n' if 'xray.tasks' in line else line

            ini_contents = ''.join(list(update()))

        logger.info('Writing %s with contents %s', working_ini_file,
                    ini_contents)

        try:
            with open(working_ini_file, 'w') as xray_ini:
                xray_ini.write(ini_contents)
        except OSError as e:
            logger.error('Failed to generate xray.ini',
                         extra={'reason': str(e)})
            raise XRayManagerError('Failed to generate xray.ini') from e
        else:
            safe_move(working_ini_file, self.ini_file())
            self.starttime = timestamp()
            logger.info('Starttime set %s', self.starttime)

    def create_ini_location(self) -> None:
        """
        Create ini location dir if it does not exist.
        Due to troubles with additional ini scan dir on DirectAdmin
        we are to create ini_location path before generating xray.ini
        """
        if not os.path.exists(self.ini_location):
            os.mkdir(self.ini_location)

    @skeleton_update
    @dbm_storage_update
    def add(self, php_version: str = None) -> None:
        """
        Add task with primary check for duplicates
        """
        if self.is_a_duplicate():
            logger.warning('Task is duplicated by URL',
                           extra={'ini_contents': self.read_ini()})
            raise XRayManagerError('Task is duplicated by URL', flag='warning')
        self.create_ini_location()
        self._add(php_version)

    @skeleton_update
    @dbm_storage_update
    def remove(self) -> None:
        """
        Remove given task form xray.ini file
        """
        working_ini_file = self.ini_file() + '.tmp'
        try:
            existing_contents = self.read_ini()
        except XRayManagerError:
            logger.error('Failed to find xray.ini',
                         extra={'xray_ini': self.ini_file()})
            return

        else:
            if ''.join(existing_contents).find(self.generate()) < 0:
                logger.error('Failed to find given task in xray.ini',
                             extra={'xray_ini': self.ini_file(),
                                    'ini_contents': self.read_ini()})
                return

            new_task_list = list()

            def update():
                """
                A generator function aimed to remove task from xray.tasks field
                """
                for line in existing_contents:
                    nonlocal new_task_list
                    if 'xray.tasks' in line:
                        new_task_list = [l.strip() for l in
                                         line[11:].split(',') if
                                         self.generate() != l.strip()]
                        yield f"xray.tasks={','.join(new_task_list)}\n"
                    else:
                        yield line

            updated_contents = ''.join(list(update()))
            if not new_task_list:
                # this indicates that there are no tasks left,
                # thus the whole xray.ini should be deleted
                logger.info('Remove xray.ini: no tasks left')
                os.unlink(self.ini_file())
            else:
                logger.info('Writing %s with contents %s',
                            working_ini_file,
                            updated_contents)
                with open(working_ini_file, 'w') as xray_ini:
                    xray_ini.write(updated_contents)
                safe_move(working_ini_file, self.ini_file())

    def recalculate_counts(self) -> int:
        """
        Recalculates remaining number of time|request_qty counts
        :return: remaining number of counts
        """
        if self.tracing_by == 'time':
            current = timestamp()
            if current < self.starttime:
                logger.error(
                    'Failed to recalculate time count: current timestamp appears to be in the past',
                    extra={'currenttime': current,
                           'starttime': self.starttime})
                raise XRayManagerError(
                    'Failed to recalculate time count: current timestamp appears to be in the past relatively to task start time')
            remaining = (current - self.starttime) // 60
            recalculated = self.tracing_count - remaining
            logger.info('Remaining time recalculated: %s left',
                        recalculated)
            return recalculated
        elif self.tracing_by == 'request_qty':
            return self.initial_count - self.request_count
        raise XRayManagerError(f'Unknown tracing marker: {self.tracing_by}')

    def erase_request_id_storage(self) -> None:
        """
        Unlink request ID file (by fake ID of tracing_task)
        """
        req_id_file = os.path.join(request_data_storage, self.fake_id)
        try:
            os.unlink(req_id_file)
        except OSError as e:
            logger.warning('Failed to unlink request_id file',
                           extra={'file': req_id_file,
                                  'err': str(e)})
